from typing import Any, Sequence
import argparse


class DangrArgparse(argparse.ArgumentParser):
    """
    Custom argument parser for the 'dangr' project.

    Provides the cli for all the arguments needed for the dangr_rt and
    allows the user to add other arguments if needed.

    In the resulting parserd arguments a `config` dictionary is
    accesible. That same dict can be recieved by any DangrAnalysis.
    """
    def __init__(self, description: str) -> None:
        super().__init__(description=description)
        self._config: dict[str, Any] = {}

        self._add_dangr_argument(
            "reverse",
            "-r",
            "--reverse",
            type=bool,
            default=None,
            help="Whether or not to reverse memory, by default this depends on the binary arch."
        )

        self._add_dangr_argument(
            "max_depth",
            "-d",
            "--max-depth",
            type=int,
            default=None,
            help="Maximum depth for backward execution."
        )

        self._add_dangr_argument(
            "timeout",
            "-t",
            "--timeout",
            type=int,
            default=None,
            help="Timeout for dangr simulation step."
        )

        self._add_dangr_argument(
            "num_finds",
            "-n",
            "--num-finds",
            type=int,
            default=None,
            help="Number of findings expected for each dangr simulation step. "
                "If this number is reached, the simulation will stop, "
                "otherwise it will excetute until there are no more active states"
        )

        self._add_dangr_argument(
            "cfg_call_depth",
            "-c",
            "--cfg-call-depth",
            type=int,
            default=None,
            help="How deep in the call stack to analyze dependencies"
        )

        self._add_dangr_argument(
            "cfg_max_steps",
            "-s",
            "--cfg-max-steps",
            type=int,
            default=None,
            help="The maximum number of basic blocks to recover forthe longest path"
                 " when constructing the cfg for the dependency analysis."
        )

        self._add_dangr_argument(
            "cfg_resolve_indirect_jumps",
            "-j",
            "--cfg-resolve-indirect-jumps",
            type=int,
            default=None,
            help="Whether to enable the indirect jump resolvers for resolving indirect jumps"
                 "for dependency analysis"
        )

    def _add_dangr_argument(self, config_key: str, *args: Any, **kwargs: Any) -> None:
        self._config[config_key] = str(kwargs.get('default'))
        super().add_argument(*args, **kwargs)

    def dangr_parse_args(
        self,
        args: Sequence[str] | None = None,
        namespace: argparse.Namespace | None = None
    ) -> argparse.Namespace:
        """
        Parses arguments, updates the config dictionary only with fixed argument values,
        and attaches config as an attribute of the returned namespace.
        """
        parsed_args = self.parse_args(args, namespace)

        for key in self._config:
            if hasattr(parsed_args, key):
                self._config[key] = getattr(parsed_args, key)

        setattr(parsed_args, "config", self._config)
        return parsed_args



from dangr_rt.dangr_analysis import DangrAnalysis
from dangr_rt.expression import Eq, And, Or, Not, Add, Mul, Sub, Div, IsMax
from dangr_rt.dangr_types import Argument
from dangr_rt.dangr_argparse import DangrArgparse
from dangr_rt.jasm_findings import JasmMatch

__all__ = [
    "DangrAnalysis", "DangrArgparse",
    "Eq", "And", "Or", "Not", "Add",
    "Mul", "Sub", "Div", "IsMax",
    "Argument", "JasmMatch"
]



from typing import Final
import angr
from dangr_rt.jasm_findings import VariableMatch
from dangr_rt.dangr_types import Argument
from dangr_rt.variables import Variable, Register, Literal, Deref

class VariableFactory:
    """
    A factory class for creating Variable objects (Register, Memory, or Literal).
    """
    REGISTER_MAP = {
        1: 'rdi',
        2: 'rsi',
        3: 'rdx',
        4: 'rcx',
        5: 'r8',
        6: 'r9',
    }

    def __init__(self, project: angr.Project) -> None:
        self.project: Final = project

    def create_from_capture(self, var: VariableMatch) -> Variable:
        """
        Creates a Variable from the structural match info.
        """
        match var.value:
            case int():
                return Literal(self.project, var.value, var.addr)
            case str():
                return Register(self.project, var.value, var.addr)
            case _:
                raise ValueError(f"Unsupported matched type: {type(var.value)} for {var.name}")

    def create_from_argument(self, argument: Argument) -> Variable:
        """
        Creates a Variable from a function argument based on its index.

        Arguments:
            argument (Argument): The function argument.

        Returns:
            Variable: The corresponding Register variable.

        Raises:
            ValueError: If the argument index does not map to a register.
        """
        norm_name = self.REGISTER_MAP.get(argument.idx)
        if norm_name is None:
            raise ValueError(f"No register for argument index {argument.idx}")

        offset = self.project.arch.get_register_offset(norm_name) # type: ignore [no-untyped-call]
        reg_name = self.project.arch.register_size_names[offset, argument.size]

        return Register(self.project, reg_name, argument.call_address)

    def create_deref(
        self,
        base: Variable,
        idx: int = 0,
        reverse: bool | None = None
    ) -> Deref:
        """
        Creates a Deref from another register 
        """
        if not isinstance(base, Register):
            raise ValueError("It's only possible to dereference registers")

        return Deref(base, idx=idx, reverse=reverse)



"""
This module defines a set of classes representing variables in symbolic execution
using the angr framework.

Classes:
- Variable: An abstract base class for variables involved in symbolic execution.
    - Register: Represents a CPU register and its symbolic representation.
    - Memory: Represents a memory location and its symbolic representation.
    - Literal: Represents a constant literal value in symbolic execution.
    - Deref: Represents a dereference operation in symbolic execution.

"""

from abc import abstractmethod, ABC
from typing import override, Final
import angr
import claripy
import archinfo

from dangr_rt.dangr_types import Address, BYTE_SIZE


class Variable(ABC):
    """
    An abstract base class representing a variable.

    This class is used to represent variables like registers, memory, or literals
    that can participate in symbolic execution.
    """

    def __init__(self, project: angr.Project, ref_addr: Address) -> None:
        self.project: Final = project
        self.ref_addr: Final = ref_addr
        self.ref_state: angr.SimState | None = None

    @abstractmethod
    def set_ref_state(self, state: angr.SimState) -> None:
        """
        Set the state asociated to the variable
        """

    @abstractmethod
    def get_expr(self) -> claripy.ast.bv.BV:
        """
        Returns an angr compatible representation of the variable
        """

    @abstractmethod
    def set_value(self, value: int) -> None:
        """
        Sets a value to the variable in the symbolic state.

        Arguments:
            state (angr.SimState): The symbolic state of the program.
            value (int): The value to set.
        """

    def evaluate(self) -> int:
        """
        Evaluates the current variable in the symbolic state where it is referenced.

        Returns:
            int: The concrete value of the variable.
        """
        if self.ref_state is None:
            raise ValueError(f"Can't evaluate {self!r} if ref_state is None")

        var_value: int = self.ref_state.solver.eval(self.get_expr(), cast_to=int)
        return var_value

    def is_concrete(self) -> bool:
        """
        Checks if the variable has a concrete value in the symbolic state
        where it is referenced

        Returns:
           bool: True if the variable is concrete.
        """
        if self.ref_state is None:
            raise ValueError(f"Can't check if {self!r} is concrete if ref_state is None")

        return self.get_expr().concrete

    @abstractmethod
    def size(self) -> int:
        """
        Returns the size of the variable in bytes
        """


class Register(Variable):
    """
    A class representing a CPU register in symbolic execution.

    Attributes:
        name (str): The name of the register (e.g., 'rax', 'ebx').
        ref_addr (Address): The address where the register is used.
    """
    def __init__(self, project: angr.Project, reg_name: str, ref_addr: Address) -> None:
        super().__init__(project, ref_addr)
        self.name: Final[str] = reg_name

    @override
    def set_ref_state(self, state: angr.SimState) -> None:
        self.ref_state = state

    @override
    def get_expr(self) -> claripy.ast.bv.BV:
        if self.ref_state is None:
            raise ValueError(f"Can't get expression of {self!r} if ref_state is None")

        return getattr(self.ref_state.regs, self.name) # type: ignore[no-any-return]

    @override
    def set_value(self, value: int) -> None:
        if self.ref_state is None:
            raise ValueError(f"Can't set value of {self!r} if ref_state is None")

        setattr(self.ref_state.regs, self.name, value)

    @override
    def size(self) -> int:
        arch = self.project.arch
        offset = arch.get_register_offset(self.name)  # type: ignore [no-untyped-call]
        possible_sizes = {int(size) for offset, size in arch.register_size_names.keys()}

        size = next(
            i for i in possible_sizes if arch.register_size_names.get((offset, i), '') == self.name
        )
        return size

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Register):
            return self.name == other.name and self.ref_addr == other.ref_addr\
                   and self.project == other.project
        return False

    def __hash__(self) -> int:
        return hash((self.project, self.name, self.ref_addr))

    def __repr__(self) -> str:
        return ('<(x) ' if self.ref_state else '<') +\
                f'Register {self.name} in {hex(self.ref_addr)}>'


class Memory(Variable):
    """
    A class representing a memory location in symbolic execution.
    """
    def __init__( # pylint: disable=(too-many-arguments)
        self,
        project: angr.Project,
        addr: int,
        size: int,
        ref_addr: Address, *,
        reverse: bool | None = None
    ) -> None:

        super().__init__(project, ref_addr)
        self._size: Final[int] = size
        self.addr: Final[Address] = addr
        self.reverse = reverse if reverse is not None else self._default_reverse()

    def _default_reverse(self) -> bool:
        return self.project.arch.memory_endness == archinfo.Endness.LE

    @override
    def get_expr(self) -> claripy.ast.bv.BV:
        if self.ref_state is None:
            raise ValueError(f"Can't get expression of {self!r} if ref_state is None")

        memory = self.ref_state.memory.load(self.addr, self.size())
        return memory.reversed if self.reverse else memory

    @override
    def set_ref_state(self, state: angr.SimState) -> None:
        self.ref_state = state

    @override
    def set_value(self, value: int) -> None:
        if self.ref_state is None:
            raise ValueError(f"Can't set value of {self!r} if ref_state is None")

        bvv_value = claripy.BVV(value, self.size()*BYTE_SIZE)

        if self.reverse:
            bvv_value = bvv_value.reversed

        self.ref_state.memory.store(self.addr, bvv_value, self.size())

    @override
    def size(self) -> int:
        return self._size

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Memory):
            return self.addr == other.addr and \
                   self._size == other._size and\
                   self.project == other.project
        return False

    def __hash__(self) -> int:
        return hash((self.project, self.addr, self.size, self.ref_addr))

    def __repr__(self) -> str:
        return ('<(x) ' if self.ref_state else '<') +\
               f'Memory ({hex(self.addr)}, {self.size()}) reference in {hex(self.ref_addr)}>'


class Literal(Variable):
    """
    A class representing a literal constant value.

    Attributes:
        value (int): The literal value.
    """
    def __init__(
        self, project: angr.Project, value: int,
        ref_addr: int,
        ) -> None:
        super().__init__(project, ref_addr)
        self.value: Final[int] = value

    @override
    def get_expr(self) -> claripy.ast.bv.BV:
        return claripy.BVV(self.value, self.size()*BYTE_SIZE)

    @override
    def set_ref_state(self, state: angr.SimState) -> None:
        pass

    @override
    def set_value(self, value: int) -> None:
        raise ValueError("Can't set a value to a Literal")

    @override
    def size(self) -> int:
        lit_block = self.project.factory.block(self.ref_addr)
        return int(next(op.size for op in lit_block.capstone.insns[0].insn.operands))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Literal):
            return self.value == other.value\
                   and self.project == other.project
        return False

    def __hash__(self) -> int:
        return hash((self.project, self.value))

    def __repr__(self) -> str:
        return f'<Literal {self.value} in {hex(self.ref_addr)}>'

class Deref(Variable):
    """
    A class representing a dereference.

    Attributes:
        value (int): The literal value.

    Right now only the following are supported dereferences:
        movl $0, (%rax)    # indirect (address is in register %rax)
        movl $0, -24(%rbp) # indirect with displacement (address = base %rbp + displacement -24)
    """
    def __init__(
        self,
        base: Register,
        idx: int = 0,
        reverse: bool | None = None
    ) -> None:

        super().__init__(base.project, base.ref_addr)
        self.base: Final[Register] = base
        self.idx: Final[int] = idx
        self.reverse: Final[bool] = reverse if reverse is not None else self._default_reverse()

    def _default_reverse(self) -> bool:
        return self.project.arch.memory_endness == archinfo.Endness.LE

    @override
    def get_expr(self) -> claripy.ast.bv.BV:
        if self.ref_state is None:
            raise ValueError(f"Can't get expression of {self!r} if ref_state is None")

        mem = self.ref_state.memory.load(
            self.base.get_expr(),
            int(self.size())
        )

        return mem.reversed if self.reverse else mem

    @override
    def set_ref_state(self, state: angr.SimState) -> None:
        self.ref_state = state
        self.base.ref_state = state

    @override
    def set_value(self, value: int) -> None:
        if self.ref_state is None:
            raise ValueError(f"Can't set value of {self!r} if ref_state is None")

        bvv_value = claripy.BVV(value, self.size()*BYTE_SIZE)
        if self.reverse:
            bvv_value = bvv_value.reversed

        self.ref_state.memory.store(self.base.get_expr(), bvv_value, self.size())

    @override
    def size(self) -> int:
        deref_block = self.project.factory.block(self.ref_addr)
        return int(deref_block.capstone.insns[0].insn.operands[0].size)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Deref):
            return self.base == other.base and self.idx == other.idx
        return False

    def __hash__(self) -> int:
        return hash((self.base, self.idx))

    def __repr__(self) -> str:
        return f'<Deref ${self.idx} + {self.base!r}>'



from typing import Final, Callable, Any
from dataclasses import dataclass
import angr

from dangr_rt.variables import Variable
from dangr_rt.dangr_types import Address, CFGNode
ConcreteState = dict['Variable', int]

def initialize_state(
    project: angr.Project,
    start: Address,
    initial_values: ConcreteState | None = None,
    add_options: set[str] | None = None
) -> angr.SimState:
    """"
    Sets the initial values in a new state and returns it
    """
    state: angr.SimState = project.factory.blank_state( # type: ignore [no-untyped-call]
        addr=start,
        add_options=add_options
    )

    if initial_values is None:
        return state

    for var, value in initial_values.items():
        var.set_ref_state(state)
        var.set_value(value)

    return state


class ForwardSimulation:
    """
    Simulate until reaching a target
    """
    def __init__(
        self,
        project: angr.Project,
        num_finds: int,
        timeout: int | None = None
    ) -> None:
        self.project = project
        self.timeout = timeout
        self.num_finds = num_finds

    def simulate(
        self,
        initial_state: angr.SimState,
        target: Address
    ) -> list[angr.SimState]:
        """
        Simulate from a given state until reaching a target address
        """
        simulation = self.project.factory.simulation_manager(initial_state)
        simulation.use_technique(
            angr.exploration_techniques.Timeout(self.timeout)
        ) # type: ignore [no-untyped-call]
        simulation.explore(find=target, num_find=self.num_finds) # type: ignore [no-untyped-call]

        return simulation.found


class HookSimulation:
    """
    Simulate until reaching a target
    """
    def __init__(
        self,
        project: angr.Project,
        init_addr: Address,
        stop: Callable[[list[angr.SimState]], bool],
        initial_values: None | ConcreteState = None,
        **inspect_kwargs: Any
    ) -> None:

        self.project = project
        self.init_addr = init_addr
        self.stop = stop
        self.initial_values = initial_values
        self.inspect_kwargs = inspect_kwargs

    def simulate(self) -> list[angr.SimState]:
        initial = initialize_state(self.project, self.init_addr, self.initial_values)
        simulation = self.project.factory.simulation_manager(initial)
        initial.inspect.b(**self.inspect_kwargs) # type: ignore [no-untyped-call]

        while simulation.active and not self.stop(simulation.active):
            simulation.step() # type: ignore [no-untyped-call]

        return simulation.active


@dataclass
class RecursiveCtx:
    current_depth: int
    backup_state: angr.SimState | None
    path: list[CFGNode]


class BackwardSimulation:
    """
    Simualte backwards until variables are concrete
    """
    _EXTERNAL_ADDR_SPACE_BASE: Final = 0x50_0000
    DEFAULT_MAX_DEPTH: Final[int] = 1

    def __init__( # pylint: disable=too-many-arguments
        self, project: angr.Project, *, target: Address,
        cfg: angr.analyses.CFGFast, variables: list[Variable],
        max_depth: int | None = None
    ) -> None:

        self.project = project
        self.target = target
        self.cfg = cfg
        self.variables = variables
        self.states_found: list[angr.SimState] = []
        self.max_depth: Final = max_depth or self.DEFAULT_MAX_DEPTH

    def simulate(self) -> list[angr.SimState]:
        target_node = self.cfg.model.get_any_node(self.target)

        if target_node is None:
            raise ValueError("Target node not found")

        rec_ctx = RecursiveCtx(0, None, [target_node])
        self._rec_simulate(target_node, rec_ctx)
        return self.states_found

    def _node_addr(self, node: CFGNode) -> Address:
        if not isinstance(node.addr, int):
            raise ValueError(f"Unsupported node {node}")
        return node.addr

    def _rec_simulate(
        self,
        target_node: CFGNode,
        rec_ctx: RecursiveCtx
    ) -> None:
        """
        Simulate the execution of the program until reaching the `function_node`, the simulation
        is done on the `path` and starts on the last node of the path.
        
        It will recursively look for the path that makes the `registers` concrete and when found
        it returns the concrete values of the registers in that path.
        """
        initial_node = rec_ctx.path[-1]
        state = self._simulate_slice(self._node_addr(initial_node), target_node, rec_ctx.path)

        if not state:
            self._handle_no_found_state(rec_ctx)
            return

        self._set_state_to_vars(state)

        if self._rec_simulation_stop(rec_ctx):
            self.states_found.append(state)
            return

        for pred in list(self.cfg.model.get_predecessors(initial_node)):
            new_rec_ctx = RecursiveCtx(rec_ctx.current_depth + 1, state, rec_ctx.path + [pred])
            self._rec_simulate(target_node, new_rec_ctx)

    def _handle_no_found_state(self, rec_ctx: RecursiveCtx) -> None:
        if rec_ctx.backup_state:
            self.states_found.append(rec_ctx.backup_state)

    def _set_state_to_vars(self, state: angr.SimState) -> None:
        for var in self.variables:
            var.set_ref_state(state)

    def _rec_simulation_stop(self, rec_ctx: RecursiveCtx) -> bool:
        happy_stop =  all(var.is_concrete() for var in self.variables)
        forced_stop = rec_ctx.current_depth >= self.max_depth
        return happy_stop or forced_stop

    def _simulate_slice(
        self,
        start: Address,
        target_node: CFGNode,
        pred: list[CFGNode],
    ) -> angr.SimState | None:

        initial = initialize_state(self.project, start)
        simgr = self.project.factory.simulation_manager(initial)
        state_found = self._get_finding(simgr, self._node_addr(target_node))

        while simgr.active and not state_found:
            self._remove_states(simgr.active, pred)
            simgr.step() # type: ignore [no-untyped-call]
            state_found = self._get_finding(simgr, self._node_addr(target_node))

        return state_found

    def _get_finding(self, simgr: angr.SimulationManager, target: Address) -> angr.SimState | None:
        return next((state for state in simgr.active if state.addr == target), None)

    def _remove_states(self, active_states: list[angr.SimState], pred: list[CFGNode]) -> None:
        for state in active_states:
            if self._remove_condition(state, pred):
                active_states.remove(state)

    def _remove_condition(self, state: angr.SimState, pred: list[CFGNode]) -> bool:
        already_visited = state.addr in state.history.bbl_addrs
        is_external_block =  state.addr >= self._EXTERNAL_ADDR_SPACE_BASE
        is_in_slice = state.addr in [p.addr for p in pred]
        return not is_in_slice or is_external_block or already_visited



"""
This is a mocked version of JASM
It will be modified and correctly implemented when
JASM has support for some functionalities I need
which is already in progres but not finished.
"""

import os
from typing import Any, cast
import json
from dataclasses import dataclass
from dangr_rt.dangr_types import Address


def load_json(file_path: str) -> dict[str, Any] | None:
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = json.load(file)
        return cast(dict[str, Any], data)
    except json.JSONDecodeError as e:
        print("Error decoding JSON:", e)
        return None
    except FileNotFoundError:
        print("File not found:", file_path)
        return None
    except IsADirectoryError:
        print("Invalid file", file_path)
        return None

@dataclass
class VariableMatch:
    name: str
    value: str
    addr: int

@dataclass
class AddressMatch:
    name: str
    value: int

# jasm returns a set of JasmMatch's
class JasmMatch:
    def __init__(
        self,
        match: dict[int, str],
        variables: list[VariableMatch],
        address_captures: list[AddressMatch]
    ) -> None:
        self._match = match
        self._variables = variables
        self._address_captures = address_captures

    @property
    def variables(self) -> list[VariableMatch]:
        """
        Returns a dict with all the variables matched
        Including
        - The variable name
        - The register/literal matched
        - The address capture in the instruction if it exists
        """
        return self._variables

    @property
    def address_captures(self) -> list[AddressMatch]:
        """
        Returns a list with all the address captures.
        The keys are the anchore's names and the value is the match
        """
        return self._address_captures

    def addrmatch_from_name(self, capture_name: str) -> AddressMatch:
        for addr_capt in self._address_captures:
            if addr_capt.name == capture_name:
                return addr_capt
        raise ValueError("Invalid capture name")

    def varmatch_from_name(self, capture_name: str) -> VariableMatch:
        for varmatch in self._variables:
            if varmatch.name == capture_name:
                return varmatch
        raise ValueError("Invalid capture name")

    @property
    def start(self) -> Address:
        return list(self._match.keys())[0]

    @property
    def end(self) -> Address:
        return list(self._match.keys())[-1]

def cast_var_value(var_value: str | int) -> int | str:
    try:
        return int(var_value, 16)
    except ValueError:
        return var_value

def _parse_jasm_output(jasm_out: list[dict[str, Any]]) -> list[JasmMatch]:
    all_matches = []
    for j_match in jasm_out:
        out_match = {}
        out_vars = []
        out_addr_capt = []
        for m in j_match['match']:
            addr, instr = m.split("::")
            out_match[int(addr, 16) + 0x40_0000] = instr

        for name_capt, info_capt in j_match['name-captures'].items():
            out_vars.append(
                VariableMatch(
                    name=name_capt,
                    value=cast_var_value(info_capt[0]),
                    addr=int(info_capt[1], 16) + 0x40_0000
                )
            )
        for addr_capt, info_capt in j_match['address-captures'].items():
            out_addr_capt.append(
                AddressMatch(
                    name=addr_capt,
                    value=int(info_capt, 16) + 0x40_0000
                ))

        all_matches.append(
            JasmMatch(
                match=out_match,
                variables=out_vars,
                address_captures=out_addr_capt
        ))

    return all_matches

constraint_bug_matches = [JasmMatch( {0x40_1144: "", 0x401147: "", 0x40_114a: ""}, [
    VariableMatch('x', 'edx', 0x40_114a),
    VariableMatch('y', 'eax', 0x40_114a)], []
)]


mock_small_bmp_support_lib = {
    '1339': [ JasmMatch(
        match={0x40_1339:''},
        address_captures=[
            AddressMatch(name="alloc_call", value=0x40_1339),
            AddressMatch(name="_target", value=0x40_1339)
        ],
        variables=[])],
    '12c5': [ JasmMatch(
        match={0x40_12c5:''},
        address_captures=[
            AddressMatch(name="alloc_call", value=0x40_12c5),
            AddressMatch(name="_target", value=0x40_12c5)
        ],
        variables=[])]
}

mock_hardware_breakpoint = {
    '11f3': [ JasmMatch(
        match={0x40_11f3:''},
        address_captures=[
            AddressMatch(name="ptrace_call", value=0x40_11f3),
            AddressMatch(name="_target", value=0x40_11f3),
        ],
        variables=[]
    )]
}

mock_software_breakpoint = {
    '0038': [JasmMatch(
        match={
            0x40_002b: "some_instruction",
            0x40_0038: "some_instruction"
        },
        address_captures=[
            AddressMatch(name='cmp-address', value=0x40_0038)
        ],
        variables=[
            VariableMatch(name="ptr",value='rax', addr=0x40_002b),
            VariableMatch('y', value=0xf223, addr=0x40_0038),
            VariableMatch(name='z', value='eax', addr=0x40_0038)
        ]
    )],
    '0002': [JasmMatch(
        match={
            0x40_0000: 'some_instruction',
            0x40_0002: 'some_instruction'
        },
        address_captures=[AddressMatch(name='cmp-address', value=0x40_0002)],
        variables=[
            VariableMatch(name='ptr', value='rdi', addr=0x40_0000),
            VariableMatch(name='y', value=0xfa1e0ff3, addr=0x40_0002),
            VariableMatch(name='z', value='eax', addr=0x40_0002)
        ],
    )],
    '0021': [JasmMatch(
        match={0x40_0014:'ignore', 0x40_0021: 'ignore'},
        address_captures=[AddressMatch(name='cmp-address', value=0x40_0021)],
        variables=[
            VariableMatch(name='ptr', value='rax', addr=0x40_0014),
            VariableMatch(name='y',   value='edx', addr=0x40_0021),
            VariableMatch(name='z',   value='eax', addr=0x40_0021)
        ]
    )],
    'd490': [JasmMatch(
        match={0x40_d48e:'', 0x40_d490:''},
        address_captures=[AddressMatch(name='cmp-address', value=0x40_d490)],
        variables=[
            VariableMatch(name='ptr', value='rdi', addr=0x40_d48e),
            VariableMatch(name='y', value='edx', addr=0x40_d490),
            VariableMatch(name='z', value='eax', addr=0x40_d490)
        ]
    )]
}

mock_uncontrolled_input = {
    '0078' : [JasmMatch(
            match={0x40_0059:'', 0x40_0078: ''},
            address_captures=[
                AddressMatch(name='deref-address', value=0x40_0078),
                AddressMatch(name='_target', value=0x40_0078),
            ],
            variables=[VariableMatch(name='ptr', value='rax', addr=0x40_0078)]
        )],
    '00ce': [JasmMatch(
            match={0x40_00be: '', 0x40_00ce: ''},
            address_captures=[AddressMatch(name='deref-address', value= 0x40_00ce)],
            variables=[VariableMatch(name='ptr', value='rax', addr=0x40_00ce)]
        )],
    '00b5': [JasmMatch(
            match={0x40_0081:'', 0x40_00b5: ''},
            address_captures=[AddressMatch(name='deref-address', value=0x40_00b5)],
            variables=[VariableMatch(name='ptr', value='rax', addr=0x40_00b5)]
        )],
    '0125': [JasmMatch(
            match={0x40_0109:'', 0x40_0125: ''},
            address_captures=[AddressMatch(name='deref-address', value=0x40_0125)],
            variables=[VariableMatch(name='ptr', value='rax', addr=0x40_0125)]
        )]
}

def _run_jasm(jasm_pattern: str, binary_path: str) -> list[JasmMatch]:
    path_to_mock = os.path.join(os.path.dirname(__file__), "jasm_mock")
    _ = binary_path
    match jasm_pattern:
        case 'software_breakpoint_pattern':
            path_to_mock = os.path.join(path_to_mock, "sw_brk_jasm_out.json")
            jasm_match_uparsed = load_json(path_to_mock)
            return _parse_jasm_output(jasm_match_uparsed)
        case 'mock constraint_bug':
            return constraint_bug_matches
        case 'mock small_bmp_support_lib_12c5':
            return mock_small_bmp_support_lib['12c5']
        case 'mock small_bmp_support_lib_1339':
            return mock_small_bmp_support_lib['1339']
        case 'mock hardware_breakpoint':
            return mock_hardware_breakpoint['11f3']
        case 'mock software_breakpoint_0038':
            return mock_software_breakpoint['0038']
        case 'mock software_breakpoint_0002':
            return mock_software_breakpoint['0002']
        case 'mock software_breakpoint_0021':
            return mock_software_breakpoint['0021']
        case 'mock software_breakpoint_d490':
            return mock_software_breakpoint['d490']
        case 'mock uncontrolled_input_0078':
            return mock_uncontrolled_input['0078']
        case 'mock uncontrolled_input_00ce':
            return mock_uncontrolled_input['00ce']
        case 'mock uncontrolled_input_00b5':
            return mock_uncontrolled_input['00b5']
        case 'mock uncontrolled_input_0125':
            return mock_uncontrolled_input['0125']
        case _:
            raise ValueError("We are still working on this! "
            "Try using 'software_breakpoint_pattern' to get a mocked answer")


class JasmAPI:
    def __init__(self) -> None:
        pass

    def run(self, binary_path: str, jasm_pattern) -> list[JasmMatch]:
        jasm_matches = _run_jasm(jasm_pattern, binary_path)
        return jasm_matches




from collections import namedtuple
from copy import deepcopy
from typing import Final
import angr

from dangr_rt.variables import Variable
from dangr_rt.simulator import ForwardSimulation, ConcreteState, initialize_state
from dangr_rt.dangr_types import Address, AngrBool
from dangr_rt.expression import Expression

CheckpointGroup = namedtuple('CheckpointGroup', ['variables', 'constraints'])

class Checkpoints(dict[Address, CheckpointGroup]):

    def add_variable(self, address: Address, variable: Variable) -> None:
        if address not in self:
            self[address] = CheckpointGroup([], [])

        self[address].variables.append(variable)

    def add_constraint(self, address: Address, constraint: Expression[AngrBool]) -> None:
        if address not in self:
            self[address] = CheckpointGroup([], [])

        self[address].constraints.append(constraint)

    def sorted(self) -> 'Checkpoints':
        """
        Return a new Checkpoints object with items sorted by the dictionary keys.
        """
        sorted_checkpoints = Checkpoints(sorted(self.items()))
        return sorted_checkpoints

    def last_address(self) -> Address | None:
        if not self:
            return None

        last_key = next(reversed(self.sorted()), None)
        return last_key


class DangrSimulation:
    DEFAULT_NUM_FINDS: Final[int] = 64

    def __init__(
        self,
        project: angr.Project,
        num_finds: int | None = None,
        timeout: int | None = None
    ) -> None:
        self.project = project
        self.simulator = ForwardSimulation(project, num_finds or self.DEFAULT_NUM_FINDS, timeout)
        self.variables: list[Variable] = []
        self.constraints: list[Expression[AngrBool]] = []

    def add_constraint(self, constraint: Expression[AngrBool]) -> None:
        self.constraints.append(constraint)
        self.variables.extend(constraint.variables)

    def remove_constraints(self) -> None:
        self.constraints = []
        self.variables = []

    def simulate(
        self,
        target: Address,
        init_addr: Address,
        initial_values: ConcreteState | None = None
    ) -> list[angr.SimState]:
        """
        Symbolic execute adding the constraints until reaching que target
        """
        checkpoints = self._create_checkpoints(init_addr, target)
        init_state = initialize_state(self.project, init_addr, initial_values)
        return self._rec_simulate(init_state, 0, checkpoints)

    def _rec_simulate(self, active_state, checkpoint_idx: int, checkpoints: Checkpoints):

        if checkpoint_idx >= len(checkpoints.items()):
            return [active_state]

        target, action_elem = list(checkpoints.sorted().items())[checkpoint_idx]
        next_starts = self.simulator.simulate(active_state, target)
        found_states = []

        for next_start in next_starts:
            self._set_state_to_vars(action_elem.variables, next_start)
            self._add_constraints_to_state(action_elem.constraints, next_start)

            if not next_start.solver.satisfiable():
                continue

            found_states.extend(
                self._rec_simulate(next_start, checkpoint_idx+1, deepcopy(checkpoints))
            )

        return found_states

    def _set_state_to_vars(self, variables: list[Variable], state: angr.SimState) -> None:
        for var in variables:
            var.set_ref_state(state)

    def _add_constraints_to_state(
        self,
        constraints: list[Expression[AngrBool]],
        state: angr.SimState
    ) -> None:
        for constraint in constraints:
            state.solver.add(constraint.get_expr())

    def _create_checkpoints(self, init_addr: Address, target: Address) -> Checkpoints:
        checkpoints = Checkpoints()
        self._create_var_checkpoints(checkpoints)
        self._create_constr_checkpoints(checkpoints, init_addr)
        self._add_target_checkpoint(checkpoints, target)
        return checkpoints.sorted()

    def _create_var_checkpoints(self, checkpoints: Checkpoints) -> None:
        for variable in self.variables:
            checkpoints.add_variable(variable.ref_addr, variable)

    def _create_constr_checkpoints(self, checkpoints: Checkpoints, default_addr: Address) -> None:
        for constraint in self.constraints:
            checkpoints.add_constraint(constraint.ref_addr or default_addr, constraint)

    def _add_target_checkpoint(self, checkpoints: Checkpoints, target: Address) -> None:
        if checkpoints.last_address() is None or\
           checkpoints.last_address() < target: # type: ignore [operator]
            checkpoints[target] = CheckpointGroup([], [])



from abc import ABC, abstractmethod
from typing import Final, Any, Sequence
import angr
from dangr_rt.jasm_findings import JasmAPI, JasmMatch, VariableMatch
from dangr_rt.dangr_types import Address, Path, AngrBool, Argument
from dangr_rt.variables import Variable, Register
from dangr_rt.variable_factory import VariableFactory
from dangr_rt.expression import Expression
from dangr_rt.simulator import ConcreteState
from dangr_rt.arguments_analyzer import ArgumentsAnalyzer
from dangr_rt.dependency_analyzer import DependencyAnalyzer
from dangr_rt.dangr_simulation import DangrSimulation


class DangrAnalysis(ABC): # pylint: disable=too-many-instance-attributes
    """
    Class that provides all the interface necesary for the analysis
    """

    def __init__(self, binary_path: Path, config: dict[str, Any]) -> None:
        """
        Here we set all the attributes that are independent from the structural finding
        """
        # general config
        self.binary_path = binary_path
        self.config = config
        self._project: Final = angr.Project(binary_path, load_options={"auto_load_libs": False})
        self._cfg: Final = self._project.analyses.CFGFast()
        self._current_function: Address | None = None

        # helper modules init
        self._simulator: DangrSimulation | None = None
        self._variable_factory = VariableFactory(self._project)
        self._dependency_analyzer = DependencyAnalyzer(
            self._project,
            call_depth=self.config.get('cfg_call_depth', None),
            max_steps=self.config.get('cfg_max_steps', None),
            resolve_indirect_jumps=self.config.get('cfg_resolve_indirect_jumps', None)
        )
        self._arguments_analyzer = ArgumentsAnalyzer(
            self._project,
            self._cfg,
            self.config.get('max_depth', None)
        )

    def _jasm_match_set(self) -> None:
        """
        Checks if self.simulator, self.jasm_match, and self.current_function
        are not None. Raises ValueError if any of them is None, indicating that
        set_finding should be called first.
        """
        if self._simulator is None or self._current_function is None:
            raise ValueError("Analysis not properly initialized. Call `set_finding()` first.")

    def _init_match_analysis(self, jasm_match: JasmMatch) -> None:
        """
        Sets the structural finding and updates the current function.

        Args:
            jasm_match (JasmMatch): The new structural finding to set.
        """
        # self._jasm_match = jasm_match

        # Restart analysis
        self._current_function = self._find_function(jasm_match)
        self._simulator = DangrSimulation(
            project=self._project,
            num_finds=self.config.get('num_finds', None),
            timeout=self.config.get('timeout', None)
        )
        self._dependency_analyzer.create_dependency_graph(self._current_function)

    def _create_var_from_capture(self, var: VariableMatch) -> Variable:
        """
        Creates a Variable from the JASM's match info.
        """
        return self._variable_factory.create_from_capture(var)

    def _create_var_from_argument(self, argument: Argument) -> Variable:
        """
        Creates a Variable from a function argument based on its index.
        """
        return self._variable_factory.create_from_argument(argument)

    def _create_deref(self, base: Variable, idx: int = 0) -> Variable:
        """
        Creates a dereference of a base register, optionally it is possible to add an index.
        """
        return self._variable_factory.create_deref(base, idx, self.config.get('reverse', None))

    def _find_function(self, jasm_match: JasmMatch) -> Address:
        """
        Gets the address of the function that contains the structural pattern matched.

        Raises:
            ValueError: If no single function contains the matched address range, 
            typically caused by the jasm pattern spanning multiple functions.
        """
        if jasm_match is None:
            raise ValueError('Structural finding not set')

        for fn in self._cfg.kb.functions.values():
            if self._finding_in_func(jasm_match, fn):
                return Address(fn.addr)

        raise ValueError('Function not found for target address: '
        f'start at {hex(jasm_match.start)}, end at {hex(jasm_match.end)}')

    def _finding_in_func(
        self, jasm_match: JasmMatch,
        fn: angr.knowledge_plugins.functions.function.Function
    ) -> bool:
        if jasm_match is None:
            raise ValueError('Structural finding not set')

        return bool(
            (fn.addr <= jasm_match.start) and\
            (jasm_match.end <= fn.addr + fn.size))

    def _get_fn_args(self) -> Sequence[Register]:
        self._jasm_match_set()
        return self._arguments_analyzer.get_fn_args(self._current_function) # type: ignore [arg-type]

    def _concretize_fn_args(self) -> list[ConcreteState]:
        """
        Returns a list with the concrete possible values of the arguments used
        in the function being analyzed

        Returns:
            list[ConcreteState]: all the possible combinations of the arguments values
        """
        self._jasm_match_set()

        concrete_args = self._arguments_analyzer.solve_arguments(
            # already checked in _jasm_match_set() ↓
            self._current_function, # type: ignore [arg-type]
            self._get_fn_args()
        )

        if not concrete_args:
            concrete_args.append(ConcreteState())

        return concrete_args

    def _simulate(
        self,
        target: Address,
        init_state: ConcreteState | None = None
    ) -> list[angr.SimState]:

        self._jasm_match_set()
        return self._simulator.simulate( # type: ignore [union-attr]
            target, self._current_function, init_state # type: ignore [arg-type]
        )

    def _add_constraint(self, constraint: Expression[AngrBool]) -> None:
        """
        Adds a constraints to the analysis
        """
        self._jasm_match_set()
        self._simulator.add_constraint(constraint) # type: ignore [union-attr]

    def _remove_constraints(self) -> None:
        """
        Remove all constraints from execution
        """
        self._jasm_match_set()
        self._simulator.remove_constraints() # type: ignore [union-attr]

    def _unconstrained_sat(
        self, target: Address,
        concrete_args: ConcreteState | None = None
    ) -> bool:
        """
        Re runs the simulation with no constraints, returns true if there
        is some satisfiable state after the simulation
        """
        self._remove_constraints()
        return self._satisfiable(self._simulate(target, concrete_args))

    def _satisfiable(self, states: list[angr.SimState]) -> bool:
        """
        Returns True if all the constraints can be satisfied at the same time
        in any of the states given.
        """
        for state in states:
            state.solver.reload_solver() # type: ignore [no-untyped-call]
            if state.solver.satisfiable():
                return True
        return False

    def _depends(self, source: Variable, target: Variable) -> bool:
        """
        Calculates dependencies of a given variable
        """
        return self._dependency_analyzer.check_dependency(source, target)

    def analyze(self, jasm_pattern) -> str | None:
        """
        Template method that performs the analysis given a jasm pattern 
        """
        jasm_matches = JasmAPI().run(self.binary_path, jasm_pattern)

        for jasm_match in jasm_matches:
            self._init_match_analysis(jasm_match)
            match_analysis_res = self._analyze_asm_match(jasm_match)

            if match_analysis_res is not None:
                return match_analysis_res

    @abstractmethod
    def _analyze_asm_match(self, jasm_match: JasmMatch) -> str | None:
        pass
